'use strict';

import React from 'react';
import PropTypes from 'prop-types';
import ReactDOM from 'react-dom';
import ScrollResponder from '../Components/ScrollResponder';
import StyleSheet from '../StyleSheet/StyleSheet';
import View from '../View/View';
import throttle from 'domkit/throttle';
import mixin from 'react-mixin';
import autobind from 'autobind-decorator';
import ScrollViewStickyHeader from './ScrollViewStickyHeader';
import './ScrollView.css';


const SCROLLVIEW = 'ScrollView';
const INNERVIEW = 'InnerScrollView';
/**
 * Component that wraps platform ScrollView while providing
 * integration with touch locking "responder" system.
 *
 * Keep in mind that ScrollViews must have a bounded height in order to work,
 * since they contain unbounded-height children into a bounded container (via
 * a scroll interaction). In order to bound the height of a ScrollView, either
 * set the height of the view directly (discouraged) or make sure all parent
 * views have bounded height. Forgetting to transfer `{flex: 1}` down the
 * view stack can lead to errors here, which the element inspector makes
 * easy to debug.
 *
 * Doesn't yet support other contained responders from blocking this scroll
 * view from becoming the responder.
 */
class ScrollView extends React.Component {
  flag = true;
  static propTypes = {
    /**
     * Controls whether iOS should automatically adjust the content inset
     * for scroll views that are placed behind a navigation bar or
     * tab bar/ toolbar. The default value is true.
     * @platform ios
     */
    automaticallyAdjustContentInsets: PropTypes.bool,
    /**
     * The amount by which the scroll view content is inset from the edges
     * of the scroll view. Defaults to `{0, 0, 0, 0}`.
     * @platform ios
     */
    // contentInset: EdgeInsetsPropType,
    /**
     * Used to manually set the starting scroll offset.
     * The default value is `{x: 0, y: 0}`.
     * @platform ios
     */
    // contentOffset: PointPropType,
    /**
     * When true, the scroll view bounces when it reaches the end of the
     * content if the content is larger then the scroll view along the axis of
     * the scroll direction. When false, it disables all bouncing even if
     * the `alwaysBounce*` props are true. The default value is true.
     * @platform ios
     */
    bounces: PropTypes.bool,
    /**
     * When true, gestures can drive zoom past min/max and the zoom will animate
     * to the min/max value at gesture end, otherwise the zoom will not exceed
     * the limits.
     * @platform ios
     */
    bouncesZoom: PropTypes.bool,
    /**
     * When true, the scroll view bounces horizontally when it reaches the end
     * even if the content is smaller than the scroll view itself. The default
     * value is true when `horizontal={true}` and false otherwise.
     * @platform ios
     */
    alwaysBounceHorizontal: PropTypes.bool,
    /**
     * When true, the scroll view bounces vertically when it reaches the end
     * even if the content is smaller than the scroll view itself. The default
     * value is false when `horizontal={true}` and true otherwise.
     * @platform ios
     */
    alwaysBounceVertical: PropTypes.bool,
    /**
     * When true, the scroll view automatically centers the content when the
     * content is smaller than the scroll view bounds; when the content is
     * larger than the scroll view, this property has no effect. The default
     * value is false.
     * @platform ios
     */
    centerContent: PropTypes.bool,
    /**
     * These styles will be applied to the scroll view content container which
     * wraps all of the child views. Example:
     *
     *   return (
     *     <ScrollView contentContainerStyle={styles.contentContainer}>
     *     </ScrollView>
     *   );
     *   ...
     *   var styles = StyleSheet.create({
     *     contentContainer: {
     *       paddingVertical: 20
     *     }
     *   });
     */
    // contentContainerStyle: StyleSheetPropType(ViewStylePropTypes),
    /**
     * A floating-point number that determines how quickly the scroll view
     * decelerates after the user lifts their finger. Reasonable choices include
     *   - Normal: 0.998 (the default)
     *   - Fast: 0.9
     * @platform ios
     */
    decelerationRate: PropTypes.number,
    /**
     * When true, the scroll view's children are arranged horizontally in a row
     * instead of vertically in a column. The default value is false.
     */
    horizontal: PropTypes.bool,
    /**
     * When true, the ScrollView will try to lock to only vertical or horizontal
     * scrolling while dragging.  The default value is false.
     * @platform ios
     */
    directionalLockEnabled: PropTypes.bool,
    /**
     * When false, once tracking starts, won't try to drag if the touch moves.
     * The default value is true.
     * @platform ios
     */
    canCancelContentTouches: PropTypes.bool,
    /**
     * Determines whether the keyboard gets dismissed in response to a drag.
     *   - 'none' (the default), drags do not dismiss the keyboard.
     *   - 'on-drag', the keyboard is dismissed when a drag begins.
     *   - 'interactive', the keyboard is dismissed interactively with the drag and moves in
     *     synchrony with the touch; dragging upwards cancels the dismissal.
     *     On android this is not supported and it will have the same behavior as 'none'.
     */
    keyboardDismissMode: PropTypes.oneOf([
      'none', // default
      'interactive',
      'on-drag',
    ]),
    /**
     * When false, tapping outside of the focused text input when the keyboard
     * is up dismisses the keyboard. When true, the scroll view will not catch
     * taps, and the keyboard will not dismiss automatically. The default value
     * is false.
     */
    keyboardShouldPersistTaps: PropTypes.bool,
    /**
     * The maximum allowed zoom scale. The default value is 1.0.
     * @platform ios
     */
    maximumZoomScale: PropTypes.number,
    /**
     * The minimum allowed zoom scale. The default value is 1.0.
     * @platform ios
     */
    minimumZoomScale: PropTypes.number,
    /**
     * Fires at most once per frame during scrolling. The frequency of the
     * events can be contolled using the `scrollEventThrottle` prop.
     */
    onScroll: PropTypes.func,
    /**
     * Called when a scrolling animation ends.
     * @platform ios
     */
    onScrollAnimationEnd: PropTypes.func,
    /**
     * When true, the scroll view stops on multiples of the scroll view's size
     * when scrolling. This can be used for horizontal pagination. The default
     * value is false.
     * @platform ios
     */
    pagingEnabled: PropTypes.bool,
    /**
     * When false, the content does not scroll.
     * The default value is true.
     * @platform ios
     */
    scrollEnabled: PropTypes.bool,
    /**
     * This controls how often the scroll event will be fired while scrolling
     * (in events per seconds). A higher number yields better accuracy for code
     * that is tracking the scroll position, but can lead to scroll performance
     * problems due to the volume of information being send over the bridge.
     * The default value is zero, which means the scroll event will be sent
     * only once each time the view is scrolled.
     * @platform ios
     */
    scrollEventThrottle: PropTypes.number,
    /**
     * The amount by which the scroll view indicators are inset from the edges
     * of the scroll view. This should normally be set to the same value as
     * the `contentInset`. Defaults to `{0, 0, 0, 0}`.
     * @platform ios
     */
    // scrollIndicatorInsets: EdgeInsetsPropType,
    /**
     * When true, the scroll view scrolls to top when the status bar is tapped.
     * The default value is true.
     * @platform ios
     */
    scrollsToTop: PropTypes.bool,
    /**
     * When true, shows a horizontal scroll indicator.
     * The default value is true.
     */
    showsHorizontalScrollIndicator: PropTypes.bool,
    /**
     * When true, shows a vertical scroll indicator.
     * The default value is true.
     */
    showsVerticalScrollIndicator: PropTypes.bool,
    /**
     * An array of child indices determining which children get docked to the
     * top of the screen when scrolling. For example, passing
     * `stickyHeaderIndices={[0]}` will cause the first child to be fixed to the
     * top of the scroll view. This property is not supported in conjunction
     * with `horizontal={true}`.
     */
    stickyHeaderIndices: PropTypes.arrayOf(PropTypes.number),
    // style: StyleSheetPropType(ViewStylePropTypes),
    /**
     * When set, causes the scroll view to stop at multiples of the value of
     * `snapToInterval`. This can be used for paginating through children
     * that have lengths smaller than the scroll view. Used in combination
     * with `snapToAlignment`.
     * @platform ios
     */
    snapToInterval: PropTypes.number,
    /**
     * When `snapToInterval` is set, `snapToAlignment` will define the relationship
     * of the the snapping to the scroll view.
     *   - `start` (the default) will align the snap at the left (horizontal) or top (vertical)
     *   - `center` will align the snap in the center
     *   - `end` will align the snap at the right (horizontal) or bottom (vertical)
     * @platform ios
     */
    snapToAlignment: PropTypes.oneOf([
      'start', // default
      'center',
      'end',
    ]),
    /**
     * Experimental: When true, offscreen child views (whose `overflow` value is
     * `hidden`) are removed from their native backing superview when offscreen.
     * This can improve scrolling performance on long lists. The default value is
     * false.
     */
    removeClippedSubviews: PropTypes.bool,
    /**
     * The current scale of the scroll view content. The default value is 1.0.
     * @platform ios
     */
    zoomScale: PropTypes.number,
    /**
     * Called when scrollable content view of the ScrollView changes.
     *
     * Handler function is passed the content width and content height as parameters:
     * `(contentWidth, contentHeight)`
     *
     * It's implemented using onLayout handler attached to the content container
     * which this ScrollView renders.
     */
    onContentSizeChange: PropTypes.func,
  };

  static defaultProps = {
    scrollEnabled: true,
  };

  state = this.scrollResponderMixinGetInitialState();
  _stickyHeaderRefs = [];
  _stickyHeaderLayoutY = [];

  /**
   * Returns a reference to the underlying scroll responder, which supports
   * operations like `scrollTo`. All ScrollView-like components should
   * implement this method so that they can be composed while providing access
   * to the underlying scroll responder's methods.
   */
  getScrollResponder() {
    return this;
  }

  getScrollableNode() {
    return ReactDOM.findDOMNode(this);
  }

  getInnerViewNode() {
    return this.refs[INNERVIEW];
  }

  setNativeProps(props) {
    if (!props) {
      return;
    }
    let scrollView = ReactDOM.findDOMNode(this.refs[SCROLLVIEW]);
    if ('scrollEnabled' in props) {
      scrollView.style.overflow = props.scrollEnabled ? 'scroll' : 'hidden';
    }
  }

  scrollTo(x?: number | { x?: number, y?: number, animated?: boolean }, y?: number, animated?: boolean) {
    this.scrollResponderScrollTo(x, y, animated);
  }

  scrollToEnd() {
    this.scrollResponderScrollToEnd();
  }

  scrollWithoutAnimationTo(y: number = 0, x: number = 0) {
    this.scrollResponderScrollTo(x, y);
  }

  handleScroll = (e) => {
    let wrappedEvent = {
      nativeEvent: {
        contentOffset: {
          x: e.target.scrollLeft,
          y: e.target.scrollTop
        },
        contentSize: {
          width: e.target.children[0].scrollWidth,
          height: e.target.children[0].scrollHeight
        },
        layoutMeasurement: {
          height: e.target.offsetHeight,
          width: e.target.offsetWidth,
        },
      },
      webEvent: e,
    };
	if(e.target.scrollLeft === 0 && e.target.className.indexOf('horizontal-scroll') === -1){
		this._checkHeaderPositions(e.target.scrollTop);
	}
    this.props.onScroll && this.props.onScroll(wrappedEvent);
  }

  _handleContentOnLayout = (e: Object) => {
    let scrollView = ReactDOM.findDOMNode(this.refs[SCROLLVIEW]);
    if (scrollView) {
      var offsetTop = scrollView.getBoundingClientRect().top;
      e.nativeEvent.layout = {
        ...e.nativeEvent.layout,
        x: scrollView.scrollLeft,
        y: scrollView.scrollTop,
      };
    }
    const { width, height } = e.nativeEvent.layout;
    this.props.onContentSizeChange && this.props.onContentSizeChange(width, height);
    if(this.flag) {
      e.nativeEvent.layout.y = offsetTop;
      this.flag = false;
    }
    this._checkHeaderPositions(e.nativeEvent.layout.y, offsetTop);
  }

  _checkHeaderPositions = (y, offsetTop) => {
    if (this._stickyHeaderRefs && this._stickyHeaderRefs.length) {
      this._stickyHeaderRefs.forEach(_ref => {
        if (_ref && _ref.checkHeaderPosition) {
          _ref.checkHeaderPosition(y, offsetTop);
        }
      })
    }
  }

  render() {
    let scrollViewStyle = [
      styles.base,
      this.props.horizontal ? { overflowY: 'hidden' } : { overflowX: 'hidden' },
      this.props.scrollEnabled ? null : { overflow: 'hidden' },
      this.props.style,
    ];
    let contentContainerStyle = [
      styles.contentContainer,
      this.props.horizontal && styles.contentContainerHorizontal,
      this.props.contentContainerStyle,
    ];

    let { children, stickyHeaderIndices } = this.props;
    const hasStickyHeaders = stickyHeaderIndices && stickyHeaderIndices.length && !this.props.horizontal;
    if (hasStickyHeaders) {
      children = React.Children.map(this.props.children, (child, idx) => {
        const stickyHeaderIndex = stickyHeaderIndices.indexOf(idx);
        if (stickyHeaderIndex >= 0) {
          return (
            <ScrollViewStickyHeader
              ref={v => this._stickyHeaderRefs[stickyHeaderIndex] = v}
              onLayout={(layout) => { this._onStickyHeaderLayout(stickyHeaderIndex, layout) }}>
              {child}
            </ScrollViewStickyHeader>
          )
        }
        return child;
      }, this)
    }

    let contentSizeChangeProps = {};
    if (this.props.onContentSizeChange) {
      contentSizeChangeProps = {
        onLayout: this._handleContentOnLayout,
      };
    }

    let contentContainer =
      <View
        ref={INNERVIEW}
        {...contentSizeChangeProps}
        style={contentContainerStyle}
        removeClippedSubviews={this.props.removeClippedSubviews}
        collapsable={false}
        className="scroll-content-container">
        {children}
      </View>;

    let alwaysBounceHorizontal =
      this.props.alwaysBounceHorizontal !== undefined ?
        this.props.alwaysBounceHorizontal :
        this.props.horizontal;

    let alwaysBounceVertical =
      this.props.alwaysBounceVertical !== undefined ?
        this.props.alwaysBounceVertical :
        !this.props.horizontal;

    let handleScroll = (...args) => {
      this.handleScroll(...args);
    };
    // if (this.props.scrollEventThrottle && this.props.onScroll && !hasStickyHeaders) {
    //   handleScroll = throttle(this.handleScroll, this.props.scrollEventThrottle);
    // }

    let props = {
      ...this.props,
      alwaysBounceHorizontal,
      alwaysBounceVertical,
      style: (scrollViewStyle: ?Array<any>),
      onTouchStart: this.scrollResponderHandleTouchStart,
      onTouchMove: this.scrollResponderHandleTouchMove,
      onTouchEnd: this.scrollResponderHandleTouchEnd,
      onScrollBeginDrag: this.scrollResponderHandleScrollBeginDrag,
      onScrollEndDrag: this.scrollResponderHandleScrollEndDrag,
      onMomentumScrollBegin: this.scrollResponderHandleMomentumScrollBegin,
      onMomentumScrollEnd: this.scrollResponderHandleMomentumScrollEnd,
      onStartShouldSetResponder: this.scrollResponderHandleStartShouldSetResponder,
      onStartShouldSetResponderCapture: this.scrollResponderHandleStartShouldSetResponderCapture,
      // onScrollShouldSetResponder: this.scrollResponderHandleScrollShouldSetResponder,
      // onScroll: handleScroll,
      onScrollShouldSetResponder: handleScroll,
      // replace onScroll in the props
      onScroll: () => { },
      onResponderGrant: this.scrollResponderHandleResponderGrant,
      onResponderTerminationRequest: this.scrollResponderHandleTerminationRequest,
      onResponderTerminate: this.scrollResponderHandleTerminate,
      onResponderRelease: this.scrollResponderHandleResponderRelease,
      onResponderReject: this.scrollResponderHandleResponderReject,
    };
    const { data, renderItem, onContentSizeChange, ...rest } = props;
    let hideHorizontalScrollIndicator = '';
    let hideVerticalScrollIndicator = '';
    if (this.props.showsHorizontalScrollIndicator === false) {

      hideHorizontalScrollIndicator = ' hide-horizontal-indicator';
    }
    if (this.props.showsVerticalScrollIndicator === false) {
      hideVerticalScrollIndicator = ' hide-vertical-indicator';
    }

    return (
      <View {...rest} ref={SCROLLVIEW} className={'scroll-view' + (this.props.horizontal ? ' horizontal-scroll' : '') +
        hideHorizontalScrollIndicator + hideVerticalScrollIndicator}>
        {contentContainer}
      </View>
    );
  }

  _onStickyHeaderLayout = (stickyHeaderIndex, layout) => {
    this._stickyHeaderLayoutY[stickyHeaderIndex] = layout;
    const prevStickyHeaderIndex = stickyHeaderIndex - 1;
    if (prevStickyHeaderIndex < this.props.stickyHeaderIndices.length
      && this._stickyHeaderRefs[prevStickyHeaderIndex]) {
      this._stickyHeaderRefs[prevStickyHeaderIndex].setNextScrollY
        && this._stickyHeaderRefs[prevStickyHeaderIndex].setNextScrollY(layout);
    }
  }
};

var isUCBrowser = /UCBrowser/i.test(navigator.userAgent);
let styles = StyleSheet.create({
  base: isUCBrowser ? {
    overflow: 'scroll',
    WebkitOverflowScrolling: 'touch',
    flex: 1,
    display: 'block',
    zIndex: 0,
  } : {
      overflow: 'scroll',
      WebkitOverflowScrolling: 'touch',
      flex: 1,
      zIndex: 0,
    },
  contentContainer: {
    width: '100%',
  },
  contentContainerHorizontal: {
    flexDirection: 'row',
    minWidth: '100%',
    width: 'auto',
    height: '100%',
  },
});

mixin(ScrollView.prototype, ScrollResponder.Mixin);
autobind(ScrollView);

ScrollView.isReactNativeComponent = true;

export default ScrollView;
